﻿using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Tetris
{
    public enum Scene
    {
        Start,
        Playing,
        Deleting,
        GameOver,
        Pause,
    }

    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        KeyboardState keyboardState;
        KeyboardState prevKeyboardState;

        // Background Image
        Texture2D backgroundImg;
        Texture2D pixel;
        SpriteFont font;

        // field, player, quartet, scene
        Scene scene;
        Field field;
        Player player;
        Quartet nextQuartet;
        Quartet activeQuartet;

        // time & threshold (for drop interval)
        float time;
        float timeForKeyStroke;
        float threshold = 10000;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // initialize objects
            scene = Scene.Start;
            player = new Player();
            field = new Field();
            activeQuartet = Quartet.GetNext(field);
            nextQuartet = Quartet.GetNext(field);

            // initialize time
            time = 0;
            timeForKeyStroke = 0;

            // initialize keyboard state
            keyboardState = Keyboard.GetState();
            prevKeyboardState = Keyboard.GetState();

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // make textures
            backgroundImg = this.Content.Load<Texture2D>("background");
            pixel = new Texture2D(GraphicsDevice, 1, 1, 0, TextureUsage.None, SurfaceFormat.Color);
            pixel.SetData<Color>(new Color[] { Color.White });

            // make fonts
            font = this.Content.Load<SpriteFont>("Arial");
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // time for moving down
            time += (float)gameTime.ElapsedGameTime.TotalMilliseconds;

            // time for key strokes (get key strokes every 200 msec)
            timeForKeyStroke += (float)gameTime.ElapsedGameTime.TotalMilliseconds;

            // keyboard state
            keyboardState = Keyboard.GetState();
            
            switch (scene) 
            {
                case Scene.Start:
                case Scene.Pause:
                    if (this.KeyStroked(Keys.Space))
                        scene = Scene.Playing;
                    break;

                case Scene.Deleting:
                    if (time > 500)
                    {
                        time = 0;
                        activeQuartet = nextQuartet;
                        nextQuartet = Quartet.GetNext(field);
                        // Delete marked blocks and slide lines
                        field.DeleteMarkedBlocks();
                        scene = Scene.Playing;
                    }
                    break;
                    
                case Scene.Playing:
                    // check pause button
                    if (this.KeyStroked(Keys.Space))
                    {
                        scene = Scene.Pause;
                        break;
                    }

                    try
                    {
                        // move down
                        if (threshold < player.Speed * time || this.KeyStroked(Keys.Down))
                        {
                            activeQuartet.MoveDown();
                            time = 0;
                        }
                    }
                    catch (CanNotMoveException)
                    {
                        try
                        {
                            foreach (Block block in activeQuartet.blocks)
                            {
                                this.field.AddBlock(block);
                            }
                            // Mark lines for delete
                            int numDeleting = field.MarkLinesForDelete();
                            if (numDeleting > 0)
                            {
                                player.UpdateStatus(numDeleting);
                                time = 0;
                                scene = Scene.Deleting;
                            }
                            else
                            {
                                activeQuartet = nextQuartet;
                                nextQuartet = Quartet.GetNext(field);
                            }
                        }
                        catch (GameOverException)
                        {
                            scene = Scene.GameOver;
                        }
                    }

                    if (this.KeyStroked(Keys.Right))
                        activeQuartet.Move(1, 0);
                    if (this.KeyStroked(Keys.Left))
                        activeQuartet.Move(-1, 0);
                    if (this.KeyStroked(Keys.Up))
                        activeQuartet.Rotate();
                    break;
                    
                case Scene.GameOver:
                    break;
            }

            // set previous keyboard state
            prevKeyboardState = keyboardState;

            base.Update(gameTime);
        }

        // check if the specified key stroked
        protected bool KeyStroked(Keys key)
        {
            if (keyboardState.IsKeyDown(key))
            {
                if (prevKeyboardState.IsKeyUp(key) || timeForKeyStroke > 200)
                {
                    timeForKeyStroke = 0;
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            graphics.GraphicsDevice.Clear(Color.CornflowerBlue);

            // Draw background
            spriteBatch.Begin(SpriteBlendMode.None);
            spriteBatch.Draw(backgroundImg, new Rectangle(0, 0, graphics.GraphicsDevice.Viewport.Width, graphics.GraphicsDevice.Viewport.Height), Color.White);
            spriteBatch.End();

            spriteBatch.Begin();

            // Draw field window
            spriteBatch.Draw(pixel, new Rectangle(273, 48, 254, 504), new Color(0, 0, 0, 224));
            spriteBatch.Draw(pixel, new Rectangle(275, 50, 250, 500), new Color(255, 255, 255, 32));

            // Draw status window
            spriteBatch.Draw(pixel, new Rectangle(58, 48, 154, 154), new Color(0, 0, 0, 224));
            spriteBatch.Draw(pixel, new Rectangle(60, 50, 150, 150), new Color(255, 255, 255, 32));

            // Draw next Quartet window
            spriteBatch.Draw(pixel, new Rectangle(588, 48, 154, 154), new Color(0, 0, 0, 224));
            spriteBatch.Draw(pixel, new Rectangle(590, 50, 150, 150), new Color(255, 255, 255, 32));

            // Draw status (Level, Lines, Score)
            player.DrawStatus(spriteBatch, font, 60, 50);
            // activeQuartet.DrawStatus(spriteBatch, font, 60, 150);  // for debug

            // Draw next Quartet
            nextQuartet.DrawAsNextQuartet(spriteBatch, pixel, 590, 50);

            // active Quartet
            activeQuartet.Draw(spriteBatch, pixel, 275, 50);

            // Draw blocks in the field
            field.DrawBlocks(spriteBatch, pixel, 275, 50);

            switch (scene)
            {
                case Scene.Start:
                    // Draw Start window
                    spriteBatch.Draw(pixel, new Rectangle(200, 150, 400, 300), new Color(255, 255, 255, 128));
                    spriteBatch.DrawString(font, "Start: Please press space button", new Vector2(210, 300), Color.Black);
                    break;
                case Scene.Pause:
                    // Draw Pause window
                    spriteBatch.Draw(pixel, new Rectangle(200, 150, 400, 300), new Color(255, 255, 255, 128));
                    spriteBatch.DrawString(font, "Pause: Please press space button for restart", new Vector2(210, 300), Color.Black);
                    break;
                case Scene.GameOver:
                    // Draw GameOver window
                    spriteBatch.Draw(pixel, new Rectangle(200, 150, 400, 300), new Color(255, 255, 255, 128));
                    spriteBatch.DrawString(font, "Game Over", new Vector2(210, 300), Color.Black);
                    break;
            }

            spriteBatch.End();
            
            base.Draw(gameTime);
        }
    }
}
