﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Tetris
{
    // Quartet Type
    public enum QuartetType
    {
        BarShaped,
        BoxShaped,
        LShaped,
        JShaped,
        ZShaped,
        SShaped,
        TShaped,
    }

    // when block can't move, raise this exception
    class CanNotMoveException : Exception { }

    // Quartet is made of 4 blocks
    class Quartet
    {
        static Random random = new Random();
        public Block[] blocks;
        protected Field field;
        protected Color color;
        protected Vector2 pos;
        protected int rotationNum;

        public Quartet(Field field)
        {
            this.field = field;
        }

        // get next Quartet
        static public Quartet GetNext(Field field)
        {
            QuartetType type = (QuartetType)random.Next(Enum.GetValues(typeof(QuartetType)).Length);
            switch (type)
            {
                case QuartetType.BarShaped:
                    return new BarQuartet(field);
                case QuartetType.BoxShaped:
                    return new BoxQuartet(field);
                case QuartetType.LShaped:
                    return new LShapedQuartet(field);
                case QuartetType.JShaped:
                    return new JShapedQuartet(field);
                case QuartetType.ZShaped:
                    return new ZShapedQuartet(field);
                case QuartetType.SShaped:
                    return new SShapedQuartet(field);
                case QuartetType.TShaped:
                    return new TShapedQuartet(field);
            }
            throw new Exception("Unknown CartetType: " + type.ToString());
        }

        // rotation matrix (this property can be override)
        virtual protected Matrix rotationMatrix
        {
            get
            {
                return Helper.createRotationR(pos.X, pos.Y);
            }
        }

        // move down
        public void MoveDown()
        {
            try
            {
                moveBlocks(0, 1);
                this.pos.Y += 1;
            }
            catch (CanNotMoveException e)
            {
                // landing
                throw e;
            }
        }

        // move
        public void Move(int x, int y)
        {
            try
            {
                moveBlocks(x, y);
                pos.X += x;
                pos.Y += y;
            }
            catch (CanNotMoveException)
            {
                // do nothing
            }
        }

        // move blocks
        protected void moveBlocks(int x, int y)
        {
            int new_x;
            int new_y;
            // check boarder
            foreach (Block block in blocks)
            {
                new_x = block.X + x;
                new_y = block.Y + y;
                if (new_x < 0 || new_x > 9 || new_y > 19)
                    throw new CanNotMoveException();
                if (this.field.HasBlock(new_x, new_y))
                    throw new CanNotMoveException();
            }
            foreach (Block block in blocks)
            {
                block.X = block.X + x;
                block.Y = block.Y + y;
            }
        }

        // rotate
        public virtual void Rotate()
        {
            try
            {
                rotateBlocks();
                // count rotation number for custom rotation logic for each QuartetType
                rotationNum++;
                if (rotationNum == 4)
                    rotationNum = 0;
            }
            catch (CanNotMoveException)
            {
                // do nothing
            }
        }

        // rotate blocks
        protected void rotateBlocks()
        {
            Vector2 pos, new_pos;
            // check boarder
            foreach (Block block in blocks)
            {
                pos = new Vector2((float)block.X, (float)block.Y);
                new_pos = Vector2.Transform(pos, rotationMatrix);
                if (new_pos.X < 0 || new_pos.X > 9 || new_pos.Y > 19)
                    throw new CanNotMoveException();
                if (this.field.HasBlock((int)new_pos.X, (int)new_pos.Y))
                    throw new CanNotMoveException();
            }
            foreach (Block block in blocks)
            {
                pos = new Vector2((float)block.X, (float)block.Y);
                new_pos = Vector2.Transform(pos, rotationMatrix);
                block.X = (int)new_pos.X;
                block.Y = (int)new_pos.Y;
            }
        }

        // draw in the game field
        public void Draw(SpriteBatch spriteBatch, Texture2D texture, int offset_x, int offset_y)
        {
            foreach (Block block in this.blocks)
            {
                block.Draw(spriteBatch, texture, offset_x, offset_y);
            }
        }

        // draw in "next quartet" window
        public void DrawAsNextQuartet(SpriteBatch spriteBatch, Texture2D texture, int offset_x, int offset_y)
        {
            foreach (Block block in this.blocks)
            {
                // Todo: offset should be refactored.
                block.Draw(spriteBatch, texture, offset_x + 40, offset_y + 15, 4, blocks[0].Y);
            }
        }

        // Draw Quartet position in status window (for debug)
        public void DrawStatus(SpriteBatch spriteBatch, SpriteFont font, int offset_x, int offset_y)
        {
            spriteBatch.DrawString(font, string.Format("Position: ({0:D}, {1:D})", (int)pos.X, (int)pos.Y), new Vector2(offset_x + 15, offset_y + 15), Color.DarkGray);
        }

    }

    // Quartet with 2 state for rotation (Bar, ZShaped, SShaped)
    abstract class TwoPhaseQuartet : Quartet
    {
        public TwoPhaseQuartet(Field field) : base(field) { }

        protected override Matrix rotationMatrix
        {
            get
            {
                if (rotationNum % 2 == 0)
                    return Helper.createRotationR(pos.X, pos.Y);
                return Helper.createRotationL(pos.X, pos.Y);
            }
        }
    }

    // Box
    class BoxQuartet : Quartet 
    {
        public BoxQuartet(Field field) : base(field)
        {
            color = Color.DarkRed;
            pos = new Vector2(4, -1);
            blocks = new Block[4]{
                new Block(4, -1, color),
                new Block(5, -1, color),
                new Block(4, 0, color),
                new Block(5, 0, color),
                };
        }

        protected override Matrix rotationMatrix
        {
            get
            {
                return Matrix.Identity;
            }
        }
    }

    // Bar
    class BarQuartet : TwoPhaseQuartet
    {
        public BarQuartet(Field field)
            : base(field)
        {
            color = Color.MediumBlue;
            pos = new Vector2(4, -1);
            blocks = new Block[4]{
                new Block(4, -3, color),
                new Block(4, -2, color),
                new Block(4, -1, color),
                new Block(4, 0, color),
                };
        }

    }

    // LShaped
    class LShapedQuartet : Quartet 
    {
        public LShapedQuartet(Field field)
            : base(field)
        {
            color = Color.Gold;
            pos = new Vector2(4, -1);
            blocks = new Block[4]{
                new Block(4, -2, color),
                new Block(4, -1, color),
                new Block(4, 0, color),
                new Block(5, 0, color),
                };
        }
    }

    // JShaped
    class JShapedQuartet : Quartet 
    {
        public JShapedQuartet(Field field)
            : base(field)
        {
            color = Color.OrangeRed;
            pos = new Vector2(5, -1);
            blocks = new Block[4]{
                new Block(5, -2, color),
                new Block(5, -1, color),
                new Block(5, 0, color),
                new Block(4, 0, color),
                };
        }
    }

    // ZShaped
    class ZShapedQuartet : TwoPhaseQuartet 
    {
        public ZShapedQuartet(Field field)
            : base(field)
        {
            color = Color.Green;
            pos = new Vector2(4, -1);
            blocks = new Block[4]{
                new Block(5, -2, color),
                new Block(5, -1, color),
                new Block(4, -1, color),
                new Block(4, 0, color),
                };
        }
    }

    // SShaped
    class SShapedQuartet : TwoPhaseQuartet 
    {
        public SShapedQuartet(Field field)
            : base(field)
        {
            color = Color.Orange;
            pos = new Vector2(4, -1);
            blocks = new Block[4]{
                new Block(4, -2, color),
                new Block(4, -1, color),
                new Block(5, -1, color),
                new Block(5, 0, color),
                };
        }
    }

    // TShaped
    class TShapedQuartet : Quartet 
    {
        public TShapedQuartet(Field field)
            : base(field)
        {
            color = Color.DeepSkyBlue;
            pos = new Vector2(4, -1);
            blocks = new Block[4]{
                new Block(4, -2, color),
                new Block(4, -1, color),
                new Block(5, -1, color),
                new Block(4, 0, color),
                };
        }
    }

}
